@@ -1,20 +1,20 @@
 from django.views.generic import DetailView, FormView, ListView
 from django.views.generic.base import View
-from django.views.generic.edit import FormMixin, CreateView, UpdateView
+from django.views.generic.edit import FormMixin, CreateView, UpdateView, DeleteView
 from django.core.urlresolvers import reverse
 from django.shortcuts import get_object_or_404, redirect
 from django.contrib import messages
-from django.utils.translation import ungettext
+from django.utils.translation import ungettext, ugettext_lazy as _
 from django.contrib.auth import REDIRECT_FIELD_NAME
-from django.db import models
 from django.http import Http404
+from django.utils.functional import cached_property
 
 from ..models import Discussion, Folder, Recipient
 from ..forms import ComposeForm, ReplyForm, FolderForm
 from ..helpers import lookup_discussions, lookup_profiles
 from .. import settings
 from ..compat import User
-from ..utils import tznow
+from ..utils import tznow, load_class
 
 from pure_pagination import Paginator
 
@@ -26,83 +26,97 @@ class DiscussionListView(ListView):
     context_object_name = 'recipient_list'
     paginator_class = Paginator
 
-    def get_queryset(self):
-        self.user = None
+    @cached_property
+    def user(self):
+        return self.request.user
 
-        if self.kwargs.get('username'):
-            username = self.kwargs.get('username')
+    @cached_property
+    def folder(self):
+        folder = None
 
-            self.user = get_object_or_404(User, username=username)
+        folder_id = self.kwargs.get('folder_id')
 
-            qs = (self.model.objects
-                  .filter(models.Q(user=self.request.user, discussion__sender=self.user) |
-                          models.Q(user=self.user, discussion__sender=self.request.user)))
-        else:
-            qs = (self.model.objects
-                  .filter(user=self.request.user))
+        if folder_id:
+            folder = get_object_or_404(Folder.objects.filter(user=self.user), pk=folder_id)
 
-        qs = (qs.exclude(status=self.model.STATUS.deleted)
-              .exclude(folder__isnull=False)
+        return folder
+
+    def get_base_queryset(self):
+        qs = (self.model.objects
+              .filter(user=self.user)
               .order_by('-discussion__updated_at', '-discussion__created_at')
               .select_related('user'))
 
         return qs
 
+    def get_queryset(self):
+        if self.folder:
+            qs = (self.get_base_queryset()
+                  .filter(folder=self.folder).exclude(status=self.model.STATUS.deleted))
+        else:
+            qs = (self.get_base_queryset().exclude(status=self.model.STATUS.deleted)
+                  .exclude(folder__isnull=False))
+
+        return qs
+
     def get_context_data(self, **kwargs):
         context = super(DiscussionListView, self).get_context_data(**kwargs)
 
         lookup_discussions(context[self.context_object_name])
         lookup_profiles(context[self.context_object_name])
 
+        context['folder_list'] = Folder.objects.all()
+        context['folder'] = self.folder
+
         return context
 
 
-class DiscussionSentView(DiscussionListView):
+class FoldersListView(ListView):
+    template_name = 'discussions/folder/list.html'
+    paginate_by = settings.PAGINATE_BY
+    model = Folder
+    context_object_name = 'folder_list'
+    paginator_class = Paginator
+
+    def get_queryset(self):
+        qs = (self.model.objects
+              .filter(user=self.request.user)
+              .select_related('user'))
+
+        return qs
+
+
+class DiscussionSentView(load_class(settings.DISCUSSION_LIST_VIEW)):
     template_name = 'discussions/sent.html'
 
     def get_queryset(self):
-        user = self.request.user
-
-        qs = (self.model.objects
-              .filter(discussion__sender=user, user=user))
-
-        qs = (qs.exclude(status=self.model.STATUS.deleted)
-              .order_by('-discussion__created_at')
-              .select_related('user'))
-
-        return qs
+        return (self.get_base_queryset()
+                .filter(discussion__sender=self.user, folder=self.folder)
+                .exclude(status=self.model.STATUS.deleted))
 
 
-class DiscussionUnreadView(DiscussionListView):
+class DiscussionUnreadView(load_class(settings.DISCUSSION_LIST_VIEW)):
     template_name = 'discussions/unread.html'
 
     def get_queryset(self):
-        user = self.request.user
-
-        qs = (self.model.objects
-              .filter(user=user))
-
-        qs = (qs.filter(status=self.model.STATUS.unread)
-              .order_by('-discussion__updated_at', '-discussion__created_at')
-              .select_related('user'))
-
-        return qs
+        return (self.get_base_queryset()
+                .filter(status=self.model.STATUS.unread, folder=self.folder))
 
 
-class DiscussionDeletedView(DiscussionListView):
+class DiscussionReadView(load_class(settings.DISCUSSION_LIST_VIEW)):
+    template_name = 'discussions/read.html'
+
+    def get_queryset(self):
+        return (self.get_base_queryset()
+                .filter(status=self.model.STATUS.read, folder=self.folder))
+
+
+class DiscussionDeletedView(load_class(settings.DISCUSSION_LIST_VIEW)):
     template_name = 'discussions/deleted.html'
 
     def get_queryset(self):
-        user = self.request.user
-
-        qs = (self.model.objects
-              .filter(user=user))
-
-        qs = (qs.filter(status=self.model.STATUS.deleted)
-              .order_by('-discussion__created_at')
-              .select_related('user'))
-
-        return qs
+        return (self.get_base_queryset()
+                .filter(status=self.model.STATUS.deleted, folder=self.folder))
 
 
 class DiscussionDetailView(DetailView, FormMixin):
@@ -124,19 +138,15 @@ class DiscussionDetailView(DetailView, FormMixin):
         if not self.is_allowed(self.request.user):
             raise Http404
 
-        recipients = Recipient.objects.filter(discussion=self.object,
-                                              user=self.request.user)
+        self.object.mark_as_read(self.request.user)
 
-        now = tznow()
-        recipients.update(read_at=now, status=Recipient.STATUS.read)
-
-        users = self.object.recipients.select_related('user')
+        recipients = self.object.recipients.all()
 
         return dict(data, **{
-            'recipients': recipients,
-            'users': users,
+            'recipient_list': recipients,
             'form': self.get_form(self.get_form_class()),
-            'message_list': self.get_messages(self.object)
+            'message_list': self.get_messages(self.object),
+            'folder_list': Folder.objects.all()
         })
 
     def get_messages(self, discussion):
@@ -163,7 +173,7 @@ class DiscussionDetailView(DetailView, FormMixin):
 
     def get_form_kwargs(self):
         return dict(super(DiscussionDetailView, self).get_form_kwargs(), **{
-            'discussion': self.get_object()
+            'discussion': self.object
         })
 
 
@@ -175,12 +185,14 @@ class MessageComposeView(FormView):
     def get_template_names(self):
         return [self.template_name, ]
 
-    def get_initial(self):
+    @cached_property
+    def recipients(self):
         recipients = None
-        initial = {}
 
         if self.request.method == 'POST':
             recipients = self.request.POST.get('recipients', None)
+            if recipients is None and self.kwargs.get('recipients'):
+                recipients = self.kwargs.get('recipients')
         elif self.kwargs.get('recipients'):
             recipients = self.kwargs.get('recipients')
 
@@ -188,9 +200,12 @@ class MessageComposeView(FormView):
             username_list = [r.strip() for r in recipients.split('+')]
             recipients = [u for u in User.objects.filter(username__in=username_list)]
 
-            self.recipients = recipients
+        return recipients
 
-            initial['to'] = recipients
+    def get_initial(self):
+        initial = {
+            'to': self.recipients
+        }
 
         return initial.copy()
 
@@ -224,7 +239,6 @@ class MessageComposeView(FormView):
     def post(self, request, *args, **kwargs):
         if self.submit_key in request.POST:
             return super(MessageComposeView, self).post(request, *args, **kwargs)
-
         return self.get(request, *args, **kwargs)
 
 
@@ -260,8 +274,11 @@ class DiscussionMoveView(DetailView, DiscussionBulkMixin):
                 List of discussion id's that should be moved.
         """
 
-        self.object = self.get_object()
-        self.get_context_data(object=self.object)
+        self.object = None
+
+        if self.kwargs.get('folder_id'):
+            self.object = self.get_object()
+            self.get_context_data(object=self.object)
 
         discussion_ids = self.request.POST.getlist('discussion_ids')
 
@@ -272,13 +289,16 @@ class DiscussionMoveView(DetailView, DiscussionBulkMixin):
         return redirect(reverse('discussions_list'))
 
 
-class DiscussionReadView(View, DiscussionBulkMixin):
+class DiscussionMarkAsReadView(View, DiscussionBulkMixin):
     http_method_names = ['post']
 
     def post(self, *args, **kwargs):
         """
         A ``POST`` to mark as read discussions.
 
+        :param unread:
+            A Boolean that if ``True`` unread discussions.
+
         POST can have the following keys:
 
             ``discussions_ids``
@@ -294,6 +314,87 @@ class DiscussionReadView(View, DiscussionBulkMixin):
             for recipient in recipients:
                 recipient.mark_as_read()
 
+        folder_id = self.kwargs.get('folder_id')
+
+        if folder_id:
+            return redirect(reverse('discussions_list', kwargs={'folder_id': folder_id}))
+
+        return redirect(reverse('discussions_list'))
+
+
+class DiscussionMarkAsUnreadView(View, DiscussionBulkMixin):
+    http_method_names = ['post']
+
+    def post(self, *args, **kwargs):
+        """
+        A ``POST`` to mark as unread discussions.
+
+        POST can have the following keys:
+
+            ``discussions_ids``
+                List of discussion id's that should be marked as unread.
+        """
+
+        discussion_ids = self.request.POST.getlist('discussion_ids')
+
+        if discussion_ids:
+            recipients = (Recipient.objects.filter(discussion__in=self.valid_ids(discussion_ids),
+                                                   user=self.request.user)
+                          .exclude(status=Recipient.STATUS.unread))
+            for recipient in recipients:
+                recipient.mark_as_unread()
+
+        folder_id = self.kwargs.get('folder_id')
+
+        if folder_id:
+            return redirect(reverse('discussions_list', kwargs={'folder_id': folder_id}))
+
+        return redirect(reverse('discussions_list'))
+
+
+class DiscussionLeaveView(View, DiscussionBulkMixin):
+    http_method_names = ['post']
+
+    def post(self, *args, **kwargs):
+        """
+        A ``POST`` to leave discussions.
+
+        POST can have the following keys:
+
+            ``discussions_ids``
+                List of discussion id's that should be deleted.
+
+            ``next``
+                String containing the URI which to redirect to after the keys are
+                removed. Redirect defaults to the inbox view.
+
+        The ``next`` value can also be supplied in the URI with ``?next=<value>``.
+
+        """
+
+        discussion_ids = self.request.POST.getlist('discussion_ids')
+        redirect_to = self.request.REQUEST.get('next', False)
+
+        if discussion_ids:
+            for pk in self.valid_ids(discussion_ids):
+                discussion = get_object_or_404(Discussion, pk=pk)
+
+                result = discussion.delete_recipient(self.request.user)
+
+                if result:
+                    message = ungettext('You have left this discussion successfully',
+                                        'You have left these discussions successfully',
+                                        len(discussion_ids))
+                    messages.success(self.request, message, fail_silently=True)
+                else:
+                    message = ungettext('You have created this discussion, you cannot leave it',
+                                        'You have created these discussions, you cannot leave them',
+                                        len(discussion_ids))
+                    messages.error(self.request, message, fail_silently=True)
+
+        if redirect_to:
+            return redirect(redirect_to)
+
         return redirect(reverse('discussions_list'))
 
 
@@ -355,12 +456,12 @@ class DiscussionRemoveView(View, DiscussionBulkMixin):
             # Send messages
             if (len(changed_message_list) > 0):
                 if undo:
-                    message = ungettext('Discussion is succesfully restored.',
-                                        'Discussions are succesfully restored.',
+                    message = ungettext('This discussion has been succesfully restored',
+                                        'These discussions have been succesfully restored',
                                         len(changed_message_list))
                 else:
-                    message = ungettext('Discussion is successfully removed.',
-                                        'Discussions are successfully removed.',
+                    message = ungettext('This discussion has been successfully removed',
+                                        'These discussions have been successfully removed',
                                         len(changed_message_list))
 
                 messages.success(self.request, message, fail_silently=True)
@@ -368,6 +469,11 @@ class DiscussionRemoveView(View, DiscussionBulkMixin):
         if redirect_to:
             return redirect(redirect_to)
 
+        folder_id = self.kwargs.get('folder_id')
+
+        if folder_id:
+            return redirect(reverse('discussions_list', kwargs={'folder_id': folder_id}))
+
         return redirect(reverse('discussions_list'))
 
 
@@ -381,9 +487,9 @@ class FolderCreateView(CreateView):
         return redirect(self.get_success_url())
 
     def get_success_url(self):
-        return reverse('discussions_folder_detail', kwargs={
-            'folder_id': self.object.pk
-        })
+        message = _('Your folder has been successfully created')
+        messages.success(self.request, message, fail_silently=True)
+        return reverse('discussions_list')
 
 
 class FolderUpdateView(UpdateView):
@@ -402,35 +508,20 @@ class FolderUpdateView(UpdateView):
         return redirect(self.get_success_url())
 
     def get_success_url(self):
-        return reverse('discussions_folder_update', kwargs={
+        message = _('You have successfully updated this folder')
+        messages.success(self.request, message, fail_silently=True)
+        return reverse('discussions_list', kwargs={
             'folder_id': self.object.pk
         })
 
 
-class FolderDetailView(DiscussionListView):
+class FolderRemoveView(DeleteView):
+    http_method_names = ['post']
     model = Folder
-    template_name = 'discussions/folder/detail.html'
-    context_object_name = 'recipient_list'
-    context_object = 'folder'
-    paginate_by = settings.PAGINATE_BY
+    pk_url_kwarg = 'folder_id'
 
-    def get_object(self):
-        obj = get_object_or_404(Folder.objects.filter(user=self.request.user),
-                                pk=self.kwargs.get('folder_id'))
-
-        self.object = obj
-
-        return obj
-
-    def get_context_data(self, **kwargs):
-        data = super(FolderDetailView, self).get_context_data(**kwargs)
-
-        data[self.context_object] = self.object
-
-        return data
+    def get_success_url(self):
+        return reverse('discussions_list')
 
     def get_queryset(self):
-        return (self.get_object()
-                .recipients
-                .order_by('-discussion__created_at')
-                .select_related('user'))
+        return self.model.objects.filter(user=self.request.user)
